@@ -9,34 +9,36 @@ import struct
 import sys
 import time
 
-from iotop import ioprio, vmstat
-from netlink import Connection, NETLINK_GENERIC, U32Attr, NLM_F_REQUEST
-from genetlink import Controller, GeNlMessage
-
 #
 # Check for requirements:
-#   o Python >= 2.5 for AF_NETLINK sockets
 #   o Linux >= 2.6.20 with I/O accounting
+#   o Python >= 2.5 or Python 2.4 + ctypes
 #
-try:
-    socket.NETLINK_ROUTE
-    python25 = True
-except AttributeError:
-    python25 = False
 
 ioaccounting = os.path.exists('/proc/self/io')
+try:
+    import ctypes
+except ImportError:
+    has_ctypes = False
+else:
+    has_ctypes = True
 
-if not python25 or not ioaccounting:
+if not ioaccounting or not has_ctypes:
     def boolean2string(boolean):
         return boolean and 'Found' or 'Not found'
     print 'Could not run iotop as some of the requirements are not met:'
-    print '- Python >= 2.5 for AF_NETLINK support:', boolean2string(python25)
     print '- Linux >= 2.6.20 with I/O accounting support ' \
              '(CONFIG_TASKSTATS, CONFIG_TASK_DELAY_ACCT, ' \
-             'CONFIG_TASK_IO_ACCOUNTING):', \
-          boolean2string(ioaccounting)
+             'CONFIG_TASK_IO_ACCOUNTING):', boolean2string(ioaccounting)
+    print '- Python >= 2.5 or Python 2.4 with the ctypes module:', \
+        boolean2string(has_ctypes)
+
     sys.exit(1)
 
+from iotop import ioprio, vmstat
+from netlink import Connection, NETLINK_GENERIC, U32Attr, NLM_F_REQUEST
+from genetlink import Controller, GeNlMessage
+
 class DumpableObject(object):
     """Base class for all objects that allows easy introspection when printed"""
     def __repr__(self):
@@ -56,37 +58,44 @@ class Stats(DumpableObject):
         ('cancelled_write_bytes', 264)
     ]
 
+    has_blkio_delay_total = False
+
     def __init__(self, task_stats_buffer):
+        sd = self.__dict__
         for name, offset in Stats.members_offsets:
             data = task_stats_buffer[offset:offset + 8]
-            setattr(self, name, struct.unpack('Q', data)[0])
+            sd[name] = struct.unpack('Q', data)[0]
 
-    def accumulate(self, other_stats, operator=sum):
-        """Returns a new Stats object built from operator(self, other_stats)"""
-        delta_stats = Stats.__new__(Stats)
-        for name, offset in Stats.members_offsets:
-            self_value = getattr(self, name)
-            other_value = getattr(other_stats, name)
-            setattr(delta_stats, name, operator((self_value, other_value)))
-        return delta_stats
+        # This is a heuristic to detect if CONFIG_TASK_DELAY_ACCT is enabled in
+        # the kernel.
+        if not Stats.has_blkio_delay_total:
+            Stats.has_blkio_delay_total = self.blkio_delay_total != 0
 
-    def delta(self, other_stats):
-        """Returns self - other_stats"""
-        def subtract((me, other)):
-            return me - other
-        return self.accumulate(other_stats, operator=subtract)
+    def accumulate(self, other_stats, destination, coeff=1):
+        """Update destination from operator(self, other_stats)"""
+        dd = destination.__dict__
+        sd = self.__dict__
+        od = other_stats.__dict__
+        for member, offset in Stats.members_offsets:
+            dd[member] = sd[member] + coeff * od[member]
+
+    def delta(self, other_stats, destination):
+        """Update destination with self - other_stats"""
+        return self.accumulate(other_stats, destination, coeff=-1)
 
     def is_all_zero(self):
+        sd = self.__dict__
         for name, offset in Stats.members_offsets:
-            if getattr(self, name) != 0:
+            if sd[name] != 0:
                 return False
         return True
 
     @staticmethod
     def build_all_zero():
         stats = Stats.__new__(Stats)
+        std = stats.__dict__
         for name, offset in Stats.members_offsets:
-            setattr(stats, name, 0)
+            std[name] = 0
         return stats
 
 #
@@ -105,11 +114,13 @@ class TaskStatsNetlink(object):
         controller = Controller(self.connection)
         self.family_id = controller.get_family_id('TASKSTATS')
 
-    def get_single_task_stats(self, pid):
-        request = GeNlMessage(self.family_id, cmd=TASKSTATS_CMD_GET,
-                              attrs=[U32Attr(TASKSTATS_CMD_ATTR_PID, pid)],
-                              flags=NLM_F_REQUEST)
-        request.send(self.connection)
+    def build_request(self, tid):
+        return GeNlMessage(self.family_id, cmd=TASKSTATS_CMD_GET,
+                           attrs=[U32Attr(TASKSTATS_CMD_ATTR_PID, tid)],
+                           flags=NLM_F_REQUEST)
+
+    def get_single_task_stats(self, thread):
+        thread.task_stats_request.send(self.connection)
         try:
             reply = self.connection.recv()
         except OSError, e:
@@ -161,19 +172,24 @@ def safe_utf8_decode(s):
 
 class ThreadInfo(DumpableObject):
     """Stats for a single thread"""
-    def __init__(self, tid):
+    def __init__(self, tid, taskstats_connection):
         self.tid = tid
         self.mark = True
         self.stats_total = None
-        self.stats_delta = None
+        self.stats_delta = Stats.__new__(Stats)
+        self.task_stats_request = taskstats_connection.build_request(tid)
 
     def get_ioprio(self):
         return ioprio.get(self.tid)
 
+    def set_ioprio(self, ioprio_class, ioprio_data):
+        return ioprio.set_ioprio(ioprio.IOPRIO_WHO_PROCESS, self.tid,
+                                 ioprio_class, ioprio_data)
+
     def update_stats(self, stats):
         if not self.stats_total:
             self.stats_total = stats
-        self.stats_delta = stats.delta(self.stats_total)
+        stats.delta(self.stats_total, self.stats_delta)
         self.stats_total = stats
 
 
@@ -216,6 +232,7 @@ class ProcessInfo(DumpableObject):
         if uid != self.uid:
             # Maybe the process called setuid()
             self.user = None
+            self.uid = uid
         return uid
 
     def get_user(self):
@@ -229,10 +246,9 @@ class ProcessInfo(DumpableObject):
 
     def get_proc_status_name(self):
         try:
-            proc_status = open('/proc/%d/status' % self.pid)
+            first_line = open('/proc/%d/status' % self.pid).readline()
         except IOError:
             return '{no such process}'
-        first_line = proc_status.readline()
         prefix = 'Name:\t'
         if first_line.startswith(prefix):
             name = first_line[6:].strip()
@@ -261,23 +277,31 @@ class ProcessInfo(DumpableObject):
         cmdline = ' '.join(parts).strip()
         return safe_utf8_decode(cmdline)
 
-    def did_some_io(self):
-        return not all(t.stats_delta.is_all_zero() for
-                                                 t in self.threads.itervalues())
+    def did_some_io(self, accumulated):
+        if accumulated:
+            return not self.stats_accum.is_all_zero()
+        for t in self.threads.itervalues():
+            if not t.stats_delta.is_all_zero():
+                return True
+        return False
 
     def get_ioprio(self):
         priorities = set(t.get_ioprio() for t in self.threads.itervalues())
         if len(priorities) == 1:
             return priorities.pop()
-        return '?'
+        return '?dif'
+
+    def set_ioprio(self, ioprio_class, ioprio_data):
+        for thread in self.threads.itervalues():
+            thread.set_ioprio(ioprio_class, ioprio_data)
 
     def ioprio_sort_key(self):
         return ioprio.sort_key(self.get_ioprio())
 
-    def get_thread(self, tid):
+    def get_thread(self, tid, taskstats_connection):
         thread = self.threads.get(tid, None)
         if not thread:
-            thread = ThreadInfo(tid)
+            thread = ThreadInfo(tid, taskstats_connection)
             self.threads[tid] = thread
         return thread
 
@@ -287,7 +311,7 @@ class ProcessInfo(DumpableObject):
             if thread.mark:
                 del self.threads[tid]
             else:
-                stats_delta = stats_delta.accumulate(thread.stats_delta)
+                stats_delta.accumulate(thread.stats_delta, stats_delta)
 
         nr_threads = len(self.threads)
         if not nr_threads:
@@ -297,7 +321,7 @@ class ProcessInfo(DumpableObject):
         stats_delta.swapin_delay_total /= nr_threads
 
         self.stats_delta = stats_delta
-        self.stats_accum = self.stats_accum.accumulate(self.stats_delta)
+        self.stats_accum.accumulate(self.stats_delta, self.stats_accum)
 
         return True
 
@@ -326,15 +350,21 @@ class ProcessList(DumpableObject):
 
     def list_tgids(self):
         if self.options.pids:
-            for pid in self.options.pids:
-                yield pid
+            return self.options.pids
 
-        pattern = '/proc/[0-9]*'
-        if not self.options.processes:
-            pattern += '/task/*'
+        tgids = os.listdir('/proc')
+        if self.options.processes:
+            return [int(tgid) for tgid in tgids if '0' <= tgid[0] <= '9']
 
-        for path in glob.iglob(pattern):
-            yield int(os.path.basename(path))
+        tids = []
+        for tgid in tgids:
+            if '0' <= tgid[0] <= '9':
+                try:
+                    tids.extend(map(int, os.listdir('/proc/' + tgid + '/task')))
+                except OSError:
+                    # The PID went away
+                    pass
+        return tids
 
     def list_tids(self, tgid):
         if not self.options.processes:
@@ -360,8 +390,8 @@ class ProcessList(DumpableObject):
             if not process:
                 continue
             for tid in self.list_tids(tgid):
-                thread = process.get_thread(tid)
-                stats = self.taskstats_connection.get_single_task_stats(tid)
+                thread = process.get_thread(tid, self.taskstats_connection)
+                stats = self.taskstats_connection.get_single_task_stats(thread)
                 if stats:
                     thread.update_stats(stats)
                     thread.mark = False
