@@ -1,6 +1,7 @@
 import curses
 import errno
 import locale
+import math
 import optparse
 import os
 import pwd
@@ -9,8 +10,11 @@ import struct
 import sys
 import time
 
-from iotop.data import find_uids, TaskStatsNetlink, ProcessList
+from iotop.data import find_uids, TaskStatsNetlink, ProcessList, Stats
+from iotop.data import ThreadInfo
 from iotop.version import VERSION
+import ioprio
+from ioprio import IoprioSetError
 
 #
 # Utility functions for the UI
@@ -19,14 +23,16 @@ from iotop.version import VERSION
 UNITS = ['B', 'K', 'M', 'G', 'T', 'P', 'E']
 
 def human_size(size):
-    for i in xrange(len(UNITS) - 1, 0, -1):
-        base = 1 << (10 * i)
-        if 2 * base < size:
-            break
+    if size > 0:
+        sign = ''
+    elif size < 0:
+        sign = '-'
+        size = -size
     else:
-        i = 0
-        base = 1
-    return '%.2f %s' % ((float(size) / base), UNITS[i])
+        return '0.00 B'
+
+    expo = int(math.log(size / 2, 2) / 10)
+    return '%s%.2f %s' % (sign, (float(size) / (1 << (10 * expo))), UNITS[expo])
 
 def format_size(options, bytes):
     if options.kilobytes:
@@ -57,6 +63,16 @@ def format_stats(options, process, duration):
     write_bytes = display_format(written_bytes, duration)
     return io_delay, swapin_delay, read_bytes, write_bytes
 
+#
+# UI Exceptions
+#
+
+class CancelInput(Exception): pass
+class InvalidInt(Exception): pass
+class InvalidPid(Exception): pass
+class InvalidTid(Exception): pass
+class InvalidIoprioData(Exception): pass
+
 #
 # The UI
 #
@@ -64,19 +80,17 @@ def format_stats(options, process, duration):
 class IOTopUI(object):
     # key, reverse
     sorting_keys = [
-        (lambda p: p.pid, False),
-        (lambda p: p.ioprio_sort_key(), False),
-        (lambda p: p.get_user(), False),
-        (lambda p: p.stats_delta.read_bytes, True),
-        (lambda p: p.stats_delta.write_bytes -
-                   p.stats_delta.cancelled_write_bytes, True),
-        (lambda p: p.stats_delta.swapin_delay_total, True),
+        (lambda p, s: p.pid, False),
+        (lambda p, s: p.ioprio_sort_key(), False),
+        (lambda p, s: p.get_user(), False),
+        (lambda p, s: s.read_bytes, True),
+        (lambda p, s: s.write_bytes - s.cancelled_write_bytes, True),
+        (lambda p, s: s.swapin_delay_total, True),
         # The default sorting (by I/O % time) should show processes doing
         # only writes, without waiting on them
-        (lambda p: p.stats_delta.blkio_delay_total or
-                   int(not(not(p.stats_delta.read_bytes or
-                               p.stats_delta.write_bytes))), True),
-        (lambda p: p.get_cmdline(), False),
+        (lambda p, s: s.blkio_delay_total or
+                      int(not(not(s.read_bytes or s.write_bytes))), True),
+        (lambda p, s: p.get_cmdline(), False),
     ]
 
     def __init__(self, win, process_list, options):
@@ -141,6 +155,110 @@ class IOTopUI(object):
         if orig_sorting_key != self.sorting_key:
             self.sorting_reverse = IOTopUI.sorting_keys[self.sorting_key][1]
 
+    # I wonder if switching to urwid for the display would be better here
+
+    def prompt_str(self, prompt, default=None, empty_is_cancel=True):
+        self.win.hline(1, 0, ord(' ') | curses.A_NORMAL, self.width)
+        self.win.addstr(1, 0, prompt, curses.A_BOLD)
+        self.win.refresh()
+        curses.echo()
+        curses.curs_set(1)
+        inp = self.win.getstr(1, len(prompt))
+        curses.curs_set(0)
+        curses.noecho()
+        if inp not in (None, ''):
+            return inp
+        if empty_is_cancel:
+            raise CancelInput()
+        return default
+
+    def prompt_int(self, prompt, default = None, empty_is_cancel = True):
+        inp = self.prompt_str(prompt, default, empty_is_cancel)
+        try:
+            return int(inp)
+        except ValueError:
+            raise InvalidInt()
+
+    def prompt_pid(self):
+        try:
+            return self.prompt_int('PID to ionice: ')
+        except InvalidInt:
+            raise InvalidPid()
+        except CancelInput:
+            raise
+
+    def prompt_tid(self):
+        try:
+            return self.prompt_int('TID to ionice: ')
+        except InvalidInt:
+            raise InvalidTid()
+        except CancelInput:
+            raise
+
+    def prompt_data(self, ioprio_data):
+        try:
+            if ioprio_data is not None:
+                inp = self.prompt_int('I/O priority data (0-7, currently %s): '
+                                      % ioprio_data, ioprio_data, False)
+            else:
+                inp = self.prompt_int('I/O priority data (0-7): ', None, False)
+        except InvalidInt:
+            raise InvalidIoprioData()
+        if inp < 0 or inp > 7:
+            raise InvalidIoprioData()
+        return inp
+
+    def prompt_set(self, prompt, display_list, ret_list, selected):
+        try:
+            selected = ret_list.index(selected)
+        except ValueError:
+            selected = -1
+        set_len = len(display_list) - 1
+        while True:
+            self.win.hline(1, 0, ord(' ') | curses.A_NORMAL, self.width)
+            self.win.insstr(1, 0, prompt, curses.A_BOLD)
+            offset = len(prompt)
+            for i, item in enumerate(display_list):
+                display = ' %s ' % item
+                if i is selected:
+                    attr = curses.A_REVERSE
+                else:
+                    attr = curses.A_NORMAL
+                self.win.insstr(1, offset, display, attr)
+                offset += len(display)
+            while True:
+                key = self.win.getch()
+                if key in (curses.KEY_LEFT, ord('l')) and selected > 0:
+                    selected -= 1
+                    break
+                elif key in (curses.KEY_RIGHT, ord('r')) and selected < set_len:
+                    selected += 1
+                    break
+                elif key in (curses.KEY_ENTER, ord('\n'), ord('\r')):
+                    return ret_list[selected]
+                elif key in (27, curses.KEY_CANCEL, curses.KEY_CLOSE,
+                             curses.KEY_EXIT, ord('q'), ord('Q')):
+                    raise CancelInput()
+
+    def prompt_class(self, ioprio_class=None):
+        prompt = 'I/O priority class: '
+        classes_prompt = ['Real-time', 'Best-effort', 'Idle']
+        classes_ret = ['rt', 'be', 'idle']
+        if ioprio_class is None:
+            ioprio_class = 2
+        inp = self.prompt_set(prompt, classes_prompt, classes_ret, ioprio_class)
+        return inp
+
+    def prompt_error(self, error = 'Error!'):
+        self.win.hline(1, 0, ord(' ') | curses.A_NORMAL, self.width)
+        self.win.insstr(1, 0, '  %s  ' % error, curses.A_REVERSE)
+        self.win.refresh()
+        time.sleep(1)
+
+    def prompt_clear(self):
+        self.win.hline(1, 0, ord(' ') | curses.A_NORMAL, self.width)
+        self.win.refresh()
+
     def handle_key(self, key):
         def toggle_accumulated():
             self.options.accumulated ^= True
@@ -151,6 +269,41 @@ class IOTopUI(object):
             self.options.processes ^= True
             self.process_list.clear()
             self.process_list.refresh_processes()
+        def ionice():
+            try:
+                if self.options.processes:
+                    pid = self.prompt_pid()
+                    exec_unit = self.process_list.get_process(pid)
+                else:
+                    tid = self.prompt_tid()
+                    exec_unit = ThreadInfo(tid,
+                                         self.process_list.taskstats_connection)
+                ioprio_value = exec_unit.get_ioprio()
+                (ioprio_class, ioprio_data) = \
+                                          ioprio.to_class_and_data(ioprio_value)
+                ioprio_class = self.prompt_class(ioprio_class)
+                if ioprio_class == 'idle':
+                    ioprio_data = 0
+                else:
+                    ioprio_data = self.prompt_data(ioprio_data)
+                exec_unit.set_ioprio(ioprio_class, ioprio_data)
+                self.process_list.clear()
+                self.process_list.refresh_processes()
+            except IoprioSetError, e:
+                self.prompt_error('Error setting I/O priority: %s' % e.err)
+            except InvalidPid:
+                self.prompt_error('Invalid process id!')
+            except InvalidTid:
+                self.prompt_error('Invalid thread id!')
+            except InvalidIoprioData:
+                self.prompt_error('Invalid I/O priority data!')
+            except InvalidInt:
+                self.prompt_error('Invalid integer!')
+            except CancelInput:
+                self.prompt_clear()
+            else:
+                self.prompt_clear()
+
         key_bindings = {
             ord('q'):
                 lambda: sys.exit(0),
@@ -172,6 +325,10 @@ class IOTopUI(object):
                 toggle_processes,
             ord('P'):
                 toggle_processes,
+            ord('i'):
+                ionice,
+            ord('I'):
+                ionice,
             curses.KEY_LEFT:
                 lambda: self.adjust_sorting_key(-1),
             curses.KEY_RIGHT:
@@ -189,20 +346,37 @@ class IOTopUI(object):
         def format(p):
             stats = format_stats(self.options, p, self.process_list.duration)
             io_delay, swapin_delay, read_bytes, write_bytes = stats
-            line = '%5d %4s %-8s %11s %11s %7s %7s ' % (
+            if Stats.has_blkio_delay_total:
+                delay_stats = '%7s %7s ' % (swapin_delay, io_delay)
+            else:
+                delay_stats = ' ?unavailable?  '
+            line = '%5d %4s %-8s %11s %11s %s' % (
                 p.pid, p.get_ioprio(), p.get_user()[:8], read_bytes,
-                write_bytes, swapin_delay, io_delay)
-            line += p.get_cmdline()
+                write_bytes, delay_stats)
+            cmdline = p.get_cmdline()
             if not self.options.batch:
-                line = line[:self.width - 1]
+                remaining_length = self.width - len(line)
+                if 2 < remaining_length < len(cmdline):
+                    len1 = (remaining_length - 1) // 2
+                    offset2 = -(remaining_length - len1 - 1)
+                    cmdline = cmdline[:len1] + '~' + cmdline[offset2:]
+            line += cmdline
+            if not self.options.batch:
+                line = line[:self.width]
             return line
 
         def should_format(p):
-            return not self.options.only or p.did_some_io()
+            return not self.options.only or \
+                   p.did_some_io(self.options.accumulated)
 
         processes = filter(should_format, self.process_list.processes.values())
         key = IOTopUI.sorting_keys[self.sorting_key][0]
-        processes.sort(key=key, reverse=self.sorting_reverse)
+        if self.options.accumulated:
+            stats_lambda = lambda p: p.stats_accum
+        else:
+            stats_lambda = lambda p: p.stats_delta
+        processes.sort(key=lambda p: key(p, stats_lambda(p)),
+                       reverse=self.sorting_reverse)
         if not self.options.batch:
             del processes[self.height - 2:]
         return map(format, processes)
@@ -232,8 +406,9 @@ class IOTopUI(object):
             sys.stdout.flush()
         else:
             self.win.erase()
-            self.win.addstr(summary)
+            self.win.addstr(summary[:self.width])
             self.win.hline(1, 0, ord(' ') | curses.A_REVERSE, self.width)
+            remaining_cols = self.width
             for i in xrange(len(titles)):
                 attr = curses.A_REVERSE
                 title = titles[i]
@@ -242,16 +417,26 @@ class IOTopUI(object):
                 if i == self.sorting_key:
                     attr |= curses.A_BOLD
                     title += self.sorting_reverse and '>' or '<'
+                title = title[:remaining_cols]
+                remaining_cols -= len(title)
                 self.win.addstr(title, attr)
-            for i in xrange(len(lines)):
+            if Stats.has_blkio_delay_total:
+                status_msg = None
+            else:
+                status_msg = ('CONFIG_TASK_DELAY_ACCT not enabled in kernel, '
+                              'cannot determine SWAPIN and IO %')
+            num_lines = min(len(lines), self.height - 2 - int(bool(status_msg)))
+            for i in xrange(num_lines):
                 try:
-                    self.win.addstr(i + 2, 0, lines[i].encode('utf-8'))
+                    self.win.insstr(i + 2, 0, lines[i].encode('utf-8'))
                 except curses.error:
                     exc_type, value, traceback = sys.exc_info()
                     value = '%s win:%s i:%d line:%s' % \
                                        (value, self.win.getmaxyx(), i, lines[i])
                     value = str(value).encode('string_escape')
                     raise exc_type, value, traceback
+            if status_msg:
+                self.win.insstr(self.height - 1, 0, status_msg, curses.A_BOLD)
             self.win.refresh()
 
 def run_iotop_window(win, options):
@@ -305,7 +490,8 @@ which the thread is running (set using the ionice command).
 
 Controls: left and right arrows to change the sorting column, r to invert the
 sorting order, o to toggle the --only option, p to toggle the --processes
-option, a to toggle the --accumulated option, q to quit, any other key to force a refresh.''' % sys.argv[0]
+option, a to toggle the --accumulated option, q to quit, any other key to force
+a refresh.''' % sys.argv[0]
 
 def main():
     locale.setlocale(locale.LC_ALL, '')
@@ -336,7 +522,7 @@ def main():
                       help='use kilobytes instead of a human friendly unit')
     parser.add_option('-t', '--time', action='store_true', dest='time',
                       help='add a timestamp on each line (implies --batch)')
-    parser.add_option('-q', '--quiet', action='count', dest='quiet',
+    parser.add_option('-q', '--quiet', action='count', dest='quiet', default=0,
                       help='suppress some lines of header (implies --batch)')
     parser.add_option('--profile', action='store_true', dest='profile',
                       default=False, help=optparse.SUPPRESS_HELP)
